/*
 * Arc4 random number generator for OpenBSD.
 * Copyright 1996 David Mazieres <dm@lcs.mit.edu>.
 *
 * Modification and redistribution in source and binary forms is
 * permitted provided that due credit is given to the author and the
 * OpenBSD project (for instance by leaving this copyright notice
 * intact).
 */

/*
 * This code is derived from section 17.1 of Applied Cryptography,
 * second edition, which describes a stream cipher allegedly
 * compatible with RSA Labs "RC4" cipher (the actual description of
 * which is a trade secret).  The same algorithm is used as a stream
 * cipher called "arcfour" in Tatu Ylonen's ssh package.
 *
 * Here the stream cipher has been modified always to include the time
 * when initializing the state.  That makes it impossible to
 * regenerate the same random sequence twice, so this can't be used
 * for encryption, but will generate good random numbers.
 *
 * RC4 is a registered trademark of RSA Laboratories.
 */

/*
 * Modified by Micha Moskal <malekith@pld.org.pl>
 * for use with kelpie (ANSI etc.)
 * changed arc4rand to arc4rand to allow compilation on
 * OpenBSD itself ...
 * $Id: arc4random.c,v 1.2 2000/11/27 20:09:43 malekith Exp $
 */

#include "config.h"

#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>

struct arc4_stream {
	byte i;
	byte j;
	byte s[256];
};

static int rs_initialized;
static struct arc4_stream rs;

static inline void arc4_init(struct arc4_stream *as)
{
	int n;

	for (n = 0; n < 256; n++)
		as->s[n] = n;
	as->i = 0;
	as->j = 0;
}

static inline void arc4_addrandom(struct arc4_stream *as, byte * dat, int datlen)
{
	int n;
	byte si;

	as->i--;
	for (n = 0; n < 256; n++) {
		as->i = (as->i + 1);
		si = as->s[as->i];
		as->j = (as->j + si + dat[n % datlen]);
		as->s[as->i] = as->s[as->j];
		as->s[as->j] = si;
	}
}

static void arc4_stir(struct arc4_stream *as)
{
	int fd;
	struct {
		struct timeval tv;
		byte rnd[128 - sizeof(struct timeval)];
	} rdat;

	gettimeofday(&rdat.tv, NULL);
	/* this is "only" a game, we don't need /dev/random, since
	 * it can take a while to read sth from there ...
	 *              -- malekith */
	fd = open("/dev/urandom", O_RDONLY); 
	/* but if it fails ... */
	if (fd < 0)
		fd = open("/dev/random", O_RDONLY);
	if (fd >= 0) {
		read(fd, rdat.rnd, sizeof(rdat.rnd));
		close(fd);
	}
	/* fd < 0?  Ah, what the heck. We'll just take whatever was on the
	 * stack... */

	arc4_addrandom(as, (void *) &rdat, sizeof(rdat));
}

static inline byte arc4_getbyte(struct arc4_stream *as)
{
	byte si, sj;

	as->i = (as->i + 1);
	si = as->s[as->i];
	as->j = (as->j + si);
	sj = as->s[as->j];
	as->s[as->i] = sj;
	as->s[as->j] = si;
	return (as->s[(si + sj) & 0xff]);
}

void arc4rand_stir()
{
	if (!rs_initialized) {
		arc4_init(&rs);
		rs_initialized = 1;
	}
	arc4_stir(&rs);
}

void arc4rand_addrandom(byte *dat, int datlen)
{
	if (!rs_initialized)
		arc4rand_stir();
	arc4_addrandom(&rs, dat, datlen);
}

/* 
 * note that we HAVE to get equal denisty on each
 * end every number ... 
 * simple getword()%x wouldn't give us this ...
 */
i32 arc4rand(i32 x)
{
	i32 tmp;
	i32 mask;
	
	if (!rs_initialized)
		arc4rand_stir();

	for (mask = 1, tmp = x; tmp; tmp >>= 1)
		mask <<= 1;
	mask--;

	do {
		tmp = (u32)arc4_getbyte(&rs);
		if (mask > 0xff)
			tmp |= (u32)arc4_getbyte(&rs) << 8;
		if (mask > 0xffff)
			tmp |= (u32)arc4_getbyte(&rs) << 16;
		if (mask > 0xffffff)
			tmp |= (u32)arc4_getbyte(&rs) << 24;
		tmp &= mask;
	} while (tmp >= x);

	return tmp;
}
