/* $Id: binfmt.c,v 1.3 2000/11/15 01:42:17 malekith Exp $ */

/* common interface to kbin files r/w */

#include "util.h"
#include "binfmt.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#define KBIN_F_READONLY	1

#define CHUNK 1024

static int kbin_read(struct kbin_file *f, void *ptr, int n)
{
	int k;

	if (f->error || !(f->flags & KBIN_F_READONLY))
		return -1;
		
	for (;;) {
		if (f->len >= f->off + n) {
			memcpy(ptr, f->buf + f->off, n);
			f->off += n;
			return 0;
		}

		/* first copy off what left */
		if (f->buf) {
			memcpy(ptr, f->buf + f->off, f->len - f->off);
			n -= (f->len - f->off);
			ptr = (byte*)ptr + (f->len - f->off);

			f->off = f->len = 0;
		} else 
			f->buf = (byte*)malloc(CHUNK);
		
		/* if such big chunk needed, read it directly */
		if (n > CHUNK) {
			while (n >= 0) {
				k = read(f->fd, ptr, n);
				if (k == -1) {
					f->error = 1;
					return -1;
				}
				k -= n;
			}
			return 0;
		}
		
		f->len = read(f->fd, f->buf, CHUNK);
		if (f->len == -1) {
			f->error = 1;
			return -1;
		}
	}
}

static int kbin_flush(struct kbin_file *f)
{
	int k, n;

	if (f->flags & KBIN_F_READONLY)
		return -1;
		
	for (n = 0; n < f->off; n += k)
		if ((k = write(f->fd, f->buf + n, f->off - n)) == -1) {
			f->error = 1;
			return -1;
		}
	f->off = 0;

	return 0;
}

int kbin_close(struct kbin_file *f)
{
	if (!(f->flags & KBIN_F_READONLY))
		if (kbin_flush(f))
			return -1;
	free(f->buf);
	free(f->name);
	if (close(f->fd))
		return -1;
	free(f);

	return 0;
}

static int kbin_write(struct kbin_file *f, const void *ptr, int n)
{
	int k;

	if (f->error || (f->flags & KBIN_F_READONLY))
		return -1;

	if (f->buf == 0)
		f->buf = (byte*)malloc(CHUNK);
		
	/* if we have left enough space in buf, put it there */
	if (f->off + n <= CHUNK) {
		memcpy(f->buf + f->off, ptr, n);
		f->off += n;
		return 0;
	}

	/* so copy as much as we can ... */
	memcpy(f->buf + f->off, ptr, CHUNK - f->off);
	ptr = (byte*)ptr + (CHUNK - f->off);
	n -= (CHUNK - f->off);
	f->off = CHUNK;

	if (kbin_flush(f))
		return -1;

	/* if large data, copy directly */
	if (n >= CHUNK) {
		while (n > 0) {
			k = write(f->fd, ptr, n);
			if (k == -1) {
				f->error = 1;
				return -1;
			}
			n -= k;
			ptr = (byte*)ptr + k;
		}
		return 0;
	}

	/* otherwise copy it to buf */
	memcpy(f->buf, ptr, n);
	f->off += n;

	return 0;
}

struct kbin_file *kbin_open(const char *fn, int ro, const char *fline)
{
	struct kbin_file *f;
	const char *msg;
	struct kbin_file_head hd;

#define oops(m) do { msg = m; goto fail; } while (0)
	
	f = NEW(struct kbin_file);
	memset(f, 0, sizeof(*f));
	
	if (ro)
		f->flags |= KBIN_F_READONLY;
		
	f->fd = open(fn, ro ? O_RDONLY : (O_WRONLY|O_CREAT|O_TRUNC), 0666);
	if (f->fd == -1) {
		perror(fn);
		free(f);
		return 0;
	}

	f->name = strdup(fn);
	
	/* now, we should validate whatever we can read it? */
	if (ro) {
		if (kbin_read(f, &hd, sizeof(hd)))
			oops("file too short to read header");
			
		if (f->buf[0] == '#' && f->buf[1] == '!') {
			for (f->off = 0; f->off < f->len; f->off++)
				if (f->buf[f->off] == '\n')
					break;
			if (f->off == f->len)
				oops("#! line too long");
			f->off++;
			if (kbin_read(f, &hd, sizeof(hd)))
				oops("file too short to read header, "
				     "after #!");
		}
		
		if (memcmp(hd.magic, KBIN_FILE_MAGIC, 4)) 
			oops("bad magic");

		if (hd.version > KBIN_VERSION_CURRENT) 
			oops("too recent version of file");

		if (hd.endian != KBIN_HOST_ENDIAN) 
			oops("NUXI problem (wrong endian)");

		if (hd.iptr_size != sizeof(iptr))
			oops("iptr_size diffrent");
			
		if (hd.real_size != sizeof(real))
			oops("real_size diffrent");

	} else {
		/* or try to init it */
		memcpy(hd.magic, KBIN_FILE_MAGIC, 4);
		hd.version = KBIN_VERSION_CURRENT;
		hd.endian = KBIN_HOST_ENDIAN;
		hd.iptr_size = sizeof(iptr);
		hd.real_size = sizeof(real);

		if (fline) {
			kbin_write(f, fline, strlen(fline));
			if (fline[strlen(fline) - 1] != '\n')
				kbin_write(f, "\n", 1);
		}
			
		if (kbin_write(f, &hd, sizeof(hd))) 
			oops("cannot write header");
	}
	
	return f;
	
fail:
	fprintf(stderr, "%s: %s\n", fn, msg);
	kbin_close(f);
	return 0;
}

int kbin_kill_tag(struct kbin_tag *t)
{
	if (t->data)
		free(t->data);
	free(t);

	return 0;
}

struct kbin_tag *kbin_get_tag(struct kbin_file *f)
{
	struct kbin_tag *r;
	struct kbin_tag_head hd;

	if (kbin_read(f, &hd, sizeof(hd))) {
		fprintf(stderr, "%s: taghd: unexpected EOF\n", f->name);
		return 0;
	}
		
	/* let's see if it's correct */
	if (memcmp(hd.magic, KBIN_TAG_MAGIC, 3)) {
		fprintf(stderr, "%s: tag: bad magic\n", f->name);
		return 0;
	}

	r = NEW(struct kbin_tag);
	r->id = hd.tag;
	r->len = hd.length;
	r->data = (byte*)malloc(r->len + 1);
	
	if (kbin_read(f, r->data, r->len)) {
		fprintf(stderr, "%s: tag: unexpected EOF\n", f->name);
		kbin_kill_tag(r);
		return 0;
	}

	return r;
}

int kbin_add_tag(struct kbin_file *f, int id, int len, const byte *data)
{
	struct kbin_tag_head hd;

	memcpy(hd.magic, KBIN_TAG_MAGIC, 3);
	hd.tag = id;
	hd.length = len;
	if (kbin_write(f, &hd, sizeof(hd)) || kbin_write(f, data, len)) {
		fprintf(stderr, "%s: write error\n", f->name);
		return -1;
	}

	return 0;
}

int kbin_fini(struct kbin_file *f)
{
	return kbin_add_tag(f, kbin_tag_eof, 0, 0) ||
	       kbin_close(f);
}
